//!CompilerOption:AddRef:System.Diagnostics.FileVersionInfo.dll

using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Media;
using ff14bot.AClasses;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ICSharpCode.SharpZipLib.Zip;
using LlamaLibrary;
using LlamaLibrary.Helpers;
using LlamaLibrary.Logging;
using TreeSharp;
using Action = System.Action;
using LogLevel = LlamaLibrary.Logging.LogLevel;
#pragma warning disable SA1307 // Accessible fields should begin with upper-case letter
#pragma warning disable SA1309 // Field names should not begin with underscore
#nullable enable

[assembly:AssemblyVersion("2.5.4.384")]

namespace PlatypusLoader
{
    public class PlatypusLoader : BotBase
    {
        public Assembly? assembly;
        private static FileVersionInfo? GetLocalVersion()
        {
            try
            {
                if (!File.Exists(ProjectAssembly)) return null;

                return FileVersionInfo.GetVersionInfo(ProjectAssembly);
            }
            catch (Exception ex)
            {
                Log.Error($"{ex}");
                return null;
            }
        }

        private const string ProjectName = "Platypus";
        private const string CompiledAssemblyName = "Platypus.dll";

        private static readonly string VersionUrl = $"https://rbplatypus.com/downloads/{Environment.Version.Major}/version.txt";
        private static readonly string DataUrl = $"https://rbplatypus.com/downloads/{Environment.Version.Major}";

        private static readonly LLogger Log = new("PlatypusLoader", Colors.LightSkyBlue);
        private static readonly string ProjectAssembly = Path.Combine(LoaderFolderName, CompiledAssemblyName);
        private static readonly object Locker = new();
        private static string _name;
        private static string _englishName;
        private static PulseFlags _pulseFlags;
        private static Action _onButtonPress;
        private static Action _start;
        private static Action _stop;
        private static Action _initialize;
        private static Action? _pulse;
        private static Action? _onShutdown;
        private static Version _version;
        private static bool _isAutonomous;
        private static bool _wantButton;
        private static bool _requiresProfile;
        private static bool debugEnabled;
        private static bool _updated;

        private Func<Composite> _getRoot;
        
        public PlatypusLoader()
        {
            Log.Information("Starting Platypus Loader");

            if (File.Exists(DotFile($".{ProjectName}_Debug")))
            {
                debugEnabled = true;
                Log.LogLevel = LogLevel.Debug;
                Log.Debug("Debug mode set to TRUE");
            }

            Log.Debug("Checking for updates...");

            lock (Locker)
            {
                if (_updated) return;

                _updated = true;
            }

            Task.Run(async () => { await Update(LoaderFolderName); }).Wait();

            try
            {
                Unblock(ProjectAssembly);
            }
            catch (Exception e)
            {
                Log.Error("Failed to unblock");
                Log.Exception(e);
            }

            Load();
        }

        public object Api { get; private set; }

        public override Composite Root => _getRoot();

        public override string Name => _name;
        
        public override bool IsAutonomous => _isAutonomous;

        public override bool WantButton => _wantButton;

        public override bool RequiresProfile => _requiresProfile;

        public override PulseFlags PulseFlags => _pulseFlags;
        
        public override string EnglishName => _englishName;

        public override void Pulse()
        {
            _pulse?.Invoke();
        }
        
        public override void OnShutdown()
        {
            _onShutdown?.Invoke();
        }

        public override Version Version => _version;
        
        private static string LoaderFolderName => GeneralFunctions.SourceDirectory().FullName;

        public static string DotFile(string filename) => Path.Combine(AppDomain.CurrentDomain.BaseDirectory, filename);

        public override void OnButtonPress()
        {
            _onButtonPress?.Invoke();
        }

        public override void Initialize()
        {
            _initialize?.Invoke();
        }

        public override void Start()
        {
            _start?.Invoke();
        }

        public override void Stop()
        {
            _stop?.Invoke();
        }

        private static void Extract(byte[] files, string directory)
        {
            using (var stream = new MemoryStream(files))
            {
                var zip = new FastZip();
                zip.ExtractZip(stream, directory, FastZip.Overwrite.Always, null, null, null, false, true);
            }
        }

        private static async Task Update(string loaderFolderName)
        {
            var data = await TryUpdate();

            if (data == null) return;

            try
            {
                Clean(loaderFolderName);
            }
            catch (Exception e)
            {
                Log.Exception(e);
            }

            try
            {
                Extract(data, loaderFolderName);
            }
            catch (Exception e)
            {
                Log.Exception(e);
            }
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool DeleteFile(string name);

        private static bool Unblock(string fileName) => DeleteFile(fileName + ":Zone.Identifier");

        private static void Clean(string directory)
        {
            foreach (FileInfo file in new DirectoryInfo(directory).GetFiles()) file.Delete();

            foreach (DirectoryInfo dir in new DirectoryInfo(directory).GetDirectories()) dir.Delete(true);
        }

        private static async Task<byte[]> TryUpdate()
        {
            try
            {
                using var client = new HttpClient();
                var stopwatch = Stopwatch.StartNew();
                FileVersionInfo currentFileVersionInfo = GetLocalVersion();
                Version currentVersion;
                string currentVersionQueryString;

                if (currentFileVersionInfo == null)
                {
                    Log.Warning("Platypus.dll is not present or is corrupted. Installing a fresh copy.");
                    currentVersion = new Version("0.0.0.0");
                    currentVersionQueryString = "0.0.0.0";
                }
                else
                {
                    currentVersion = new Version(string.Format("{0}.{1}.{2}.{3}", currentFileVersionInfo.FileMajorPart, currentFileVersionInfo.FileMinorPart, currentFileVersionInfo.FileBuildPart, currentFileVersionInfo.FilePrivatePart));
                    currentVersionQueryString = string.Format("{0}.{1}.{2}.{3}", currentFileVersionInfo.FileMajorPart, currentFileVersionInfo.FileMinorPart, currentFileVersionInfo.FileBuildPart, currentFileVersionInfo.FilePrivatePart);
                }

                var latestVersionString = await client.GetStringAsync($"{VersionUrl}?ver={currentVersionQueryString}");

                Version latestVersion;

                try
                {
                    Log.Debug($"Local: {currentFileVersionInfo} | Latest: {latestVersionString}");

                    latestVersion = new Version(latestVersionString);

                    if (latestVersion <= currentVersion)
                    {
                        // No need to update
                        return null;
                    }
                }
                catch (Exception ex)
                {
                    Log.Error($"Unable to verify version: {ex}");
                    return null;
                }

                Log.Information($"{ProjectName} local version is {currentVersion}. Updating to {latestVersion}.");

                using HttpResponseMessage response = await client.GetAsync($"{DataUrl}/Platypus_{latestVersion.Major}.{latestVersion.Minor}.{latestVersion.Build}.{latestVersion.MinorRevision}.zip");

                if (!response.IsSuccessStatusCode)
                {
                    Log.Warning($"[Error] Could not download {ProjectName}: {response.StatusCode}");
                    return null;
                }

                using Stream inputStream = await response.Content.ReadAsStreamAsync();
                using var memoryStream = new MemoryStream();
                await inputStream.CopyToAsync(memoryStream);

                stopwatch.Stop();
                Log.Information($"Download took {stopwatch.ElapsedMilliseconds} ms.");

                return memoryStream.ToArray();
            }
            catch (Exception e)
            {
                Log.Error($"[Error] {e}");
                return null;
            }
        }

        private static Assembly? LoadAssembly(string path)
        {
            if (!File.Exists(path)) return null;

            Assembly? assembly = null;
            try
            {
                assembly = Assembly.LoadFrom(path);
            }
            catch (Exception e)
            {
                Logging.WriteException(e);
            }

            return assembly;
        }

        private void Load()
        {
            Log.Debug("Starting constructor");

            RedirectAssembly();

            Log.Debug("Redirected assemblies");

            if (!File.Exists(ProjectAssembly))
            {
                Log.Error($"Can't find {ProjectAssembly}");
                return;
            }

            Assembly assembly = LoadAssembly(ProjectAssembly);
            if (assembly == null) return;

            Log.Debug($"{assembly.GetName().Name} v{assembly.GetName().Version} loaded");
            Type baseType;
            try
            {
                baseType = assembly.DefinedTypes.FirstOrDefault(i => typeof(ICompiledBotbase).IsAssignableFrom(i));
            }
            catch (ReflectionTypeLoadException ex)
            {
                var sb = new StringBuilder();
                foreach (Exception exSub in ex.LoaderExceptions)
                {
                    sb.AppendLine(exSub.Message);
                    var exFileNotFound = exSub as FileNotFoundException;
                    if (exFileNotFound != null && !string.IsNullOrEmpty(exFileNotFound.FusionLog))
                    {
                        sb.AppendLine("Fusion Log:");
                        sb.AppendLine(exFileNotFound.FusionLog);
                    }

                    sb.AppendLine();
                }

                var errorMessage = sb.ToString();
                Log.Error(errorMessage);
                return;

                // Display or log the error based on your application.
            }
            catch (Exception e)
            {
                Log.Error("Other Exception");
                Log.Exception(e);
                return;
            }

            if (baseType == null)
            {
                Log.Error("Base type is null");
                return;
            }

            ICompiledAPIBotbase compiledAsyncBotbase;
            try
            {
                compiledAsyncBotbase = (ICompiledAPIBotbase)Activator.CreateInstance(baseType);
            }
            catch (ReflectionTypeLoadException ex)
            {
                var sb = new StringBuilder();
                foreach (Exception exSub in ex.LoaderExceptions)
                {
                    sb.AppendLine(exSub.Message);
                    var exFileNotFound = exSub as FileNotFoundException;
                    if (exFileNotFound != null && !string.IsNullOrEmpty(exFileNotFound.FusionLog))
                    {
                        sb.AppendLine("Fusion Log:");
                        sb.AppendLine(exFileNotFound.FusionLog);
                    }

                    sb.AppendLine();
                }

                var errorMessage = sb.ToString();
                Log.Error(errorMessage);
                return;

                // Display or log the error based on your application.
            }
            catch (Exception e)
            {
                Log.Error("Other Exception2");
                Log.Exception(e);
                return;
            }

            _name = compiledAsyncBotbase.Name;
            _englishName = typeof(BotBase).GetProperty("EnglishName")?.GetValue(compiledAsyncBotbase) as string;
            _pulse = baseType.GetMethod("Pulse")?.CreateDelegate(typeof(Action), compiledAsyncBotbase) as Action;
            _onShutdown = baseType.GetMethod("OnShutdown")?.CreateDelegate(typeof(Action), compiledAsyncBotbase) as Action;
            _version =  typeof(BotBase).GetProperty("Version")?.GetValue(compiledAsyncBotbase) as Version;
            _start = compiledAsyncBotbase.Start;
            _stop = compiledAsyncBotbase.Stop;
            _onButtonPress = compiledAsyncBotbase.OnButtonPress;
            _pulseFlags = compiledAsyncBotbase.PulseFlags;
            _wantButton = compiledAsyncBotbase.WantButton;
            _requiresProfile = compiledAsyncBotbase.RequiresProfile;
            _isAutonomous = compiledAsyncBotbase.IsAutonomous;
            _getRoot = compiledAsyncBotbase.GetRoot;
            _initialize = compiledAsyncBotbase.Initialize;
            Api = compiledAsyncBotbase.Api;
        }
        
        private void RedirectAssembly()
        {
            AssemblyProxy.Init();
            AppDomain.CurrentDomain.AppendPrivatePath(LoaderFolderName);
            AssemblyProxy.AddAssembly("Microsoft.Bcl.AsyncInterfaces", LoadAssembly(Path.Combine(Clio.Utilities.Utilities.AssemblyDirectory, "Microsoft.Bcl.AsyncInterfaces.dll")));       
        }
    }
}
